{********************************************************}
{                                                        }
{       How to Implement Sequential List in database     }
{                                                        }
{   Copyright (c) 2002 N. Adesh Jain, Bangalore, India   }
{                                                        }
{     All Rights Reserved.  E-Mail: adeshjain@37.com     }
{                                                        }
{********************************************************}
unit dmFloatIns;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, DBTables, Math;

type
  TdmFloatInsert = class(TDataModule)
    dsEmployee: TDataSource;
    qryEmployee: TQuery;
    qryExecSql: TQuery;
    qryEmployeeRecPosNo: TFloatField;
    qryEmployeeEmpNo: TIntegerField;
    qryEmployeeLastName: TStringField;
    qryEmployeeFirstName: TStringField;
    qryEmployeePhoneExt: TStringField;
    qryEmployeeHireDate: TDateTimeField;
    qryEmployeeSalary: TFloatField;
    procedure qryEmployeeAfterInsert(DataSet: TDataSet);
    procedure qryEmployeeBeforeInsert(DataSet: TDataSet);
  private
    { Private declarations }
    PredRecPos, SuccRecPos, CurrRecPos: Double;
    CurrRecPosStr: String;

    function getInBewteenfFloat(FirstVal, SecondVal: Double): Double;
  public
    { Public declarations }
  end;

var
  dmFloatInsert: TdmFloatInsert;

implementation

{$R *.DFM}

function getDeciFrac(aForValue: Double): String;
var
  tmpStr: String;
  i, DeciPos: SmallInt;
begin
  tmpStr := FloatToStr(aForValue);
  DeciPos := Pos('.', tmpStr);

  if DeciPos>0 then tmpStr := Copy(tmpStr, DeciPos+1, Length(tmpStr))
  else tmpStr := '';

  Result := '';
  for i:=1 to (Length(tmpStr)-1) do begin
    Result := Result + '0';
  end;

  Result := '0.'+Result;
end;

function TdmFloatInsert.getInBewteenfFloat(FirstVal, SecondVal: Double): Double;
var
  DeciFracStr: String;
  tmpFloat: Double;
  runCount: SmallInt;
begin
  {
  Note: (IMO) Insert creates room above the selected record.
              Append creates room after the last record.
  }
  Result := 0;
  runCount := 1;

  //If No Records in the table
  if ((FirstVal=0) and (SecondVal=0)) then begin
    Result := 1;
    Exit;
  end;

  //In any condition both First & Second Values are 0
  if ((FirstVal=SecondVal) or
      ((FirstVal=0) and (SecondVal<=1))) then begin
    MessageDlg('Cannot be inserted !',mtError,[mbOK],0);
    Exit;
  end;

  //Second Value is 0 & Appending Record
  if (SecondVal=0) then begin
    Result := Trunc(FirstVal)+1;
    Exit;
  end;

  if (FirstVal>0) then DeciFracStr := getDeciFrac(FirstVal)
  else DeciFracStr := getDeciFrac(SecondVal);

  while runCount<=15 do begin
    if (FirstVal>0) then begin
      tmpFloat := FirstVal + StrToFloat(DeciFracStr + '1');
    end else begin
      tmpFloat := SecondVal - StrToFloat(DeciFracStr + '1');
    end;

    DeciFracStr := DeciFracStr + '0';

    {
    You cannot compare floating point numbers (as Real) against an exact value.
    There are always errors due to loss of precision, caused by an limited
    number of binary digits (bits) used to represent a decimal value.
    }

    tmpFloat := StrToFloat(FloatToStr(tmpFloat));

    if (tmpFloat>FirstVal) and (tmpFloat<SecondVal) then begin
      Result := tmpFloat;
      Exit;
    end;

    inc(runCount);
  end;

  //because of 15 digits restriction
  if (Length(FloatToStr(Result))>16) then Result := 0;

  if (Result<=0) then begin
    Result := 0;
    MessageDlg('Cannot be inserted !',mtError,[mbOK],0);
  end;
end;

procedure TdmFloatInsert.qryEmployeeAfterInsert(DataSet: TDataSet);
begin
  qryEmployeeRecPosNo.AsString := CurrRecPosStr;

  with QryExecSql do begin
    Close;
    Sql.Clear;
    Sql.Add('Select Max(EmpNo) MaxEmpNo from Employee ');
    Open;

    qryEmployeeEmpNo.AsInteger := FieldByName('MaxEmpNo').AsInteger + 1;
  end;
end;

procedure TdmFloatInsert.qryEmployeeBeforeInsert(DataSet: TDataSet);
begin
  with QryExecSql do begin
    Close;
    Sql.Clear;
    Sql.Add('Select Count(*) NoOfRec from Employee ');
    Sql.Add('where RecPosNo>'+qryEmployeeRecPosNo.AsString+' ');
    Open;

    if (FieldByName('NoOfRec').AsInteger>0) then begin
      SuccRecPos := qryEmployeeRecPosNo.AsFloat;

      Close;
      Sql.Clear;
      Sql.Add('Select RecPosNo from Employee ');
      Sql.Add('where RecPosNo<'+qryEmployeeRecPosNo.AsString+' ');
      Sql.Add('Order by RecPosNo Desc ');
      Open;
      PredRecPos := FieldByName('RecPosNo').AsFloat;
      Close;
    end else begin
      SuccRecPos := 0;

      PredRecPos := qryEmployeeRecPosNo.AsFloat;
    end;

    Close;
  end;

  CurrRecPos := getInBewteenfFloat(PredRecPos, SuccRecPos);

  if (CurrRecPos<=0) then Abort;

  CurrRecPosStr := FloatToStr(CurrRecPos);

  with QryExecSql do begin
    Close;
    Sql.Clear;
    Sql.Add('Select Count(*) NoOfRec from Employee ');
    Sql.Add('where RecPosNo='+CurrRecPosStr+' ');
    Open;

    if (FieldByName('NoOfRec').AsInteger>0) then Abort;

    Close;
  end;
end;

end.
